{-| Module      : Graph.Tools
Description :
Copyright   : (c) CNRS, Alexandre Delanoë
License     : AGPL + CECILL v3
Maintainer  : alexandre+dev@delanoe.org
Stability   : experimental
Portability : POSIX


-}

{-# LANGUAGE ConstraintKinds #-}
{-# LANGUAGE InstanceSigs    #-}
{-# LANGUAGE NoImplicitPrelude       #-}


module Graph.Tools.Format.CSV where

import Data.Csv
import Data.Text (pack, splitOn, unpack)
import Data.Vector hiding (map, uniq)
import Prelude (read)
import Protolude
import Graph.Types
import Graph.Tools.Clean (uniq')
import qualified Eigen.SparseMatrix as SMatrix
import qualified Data.Map                          as Map
import qualified Data.ByteString.Lazy              as BL
import qualified Data.Graph.Inductive              as DGI
import qualified Data.Graph.Inductive.PatriciaTree as DGIP
import qualified Data.List                         as List
import qualified Data.Set                          as Set
import qualified Data.Vector                       as Vector

------------------------------------------------------------------------
data FileGraph = CillexGraph
               | CSV_Col Int

readFileGraph :: FileGraph -> FilePath -> IO (DGIP.Gr [Text] Double)
readFileGraph CillexGraph fp = toGraph <$> snd <$> readFileCsvCillex fp
readFileGraph (CSV_Col 2) fp = toGraph <$> snd <$> readFileCsv_Col_2   fp
readFileGraph (CSV_Col 3) fp = toGraph <$> snd <$> readFileCsv_Col_3   fp
readFileGraph (CSV_Col _) _ = panic "Not implemented"

toGraph :: ToNode a => Vector a -> (DGIP.Gr [Text] Double)
toGraph vs = DGI.mkGraph ns (uniq' $ List.concat es)
  where
    (ns,es) = List.unzip
            $ Vector.toList
            $ Vector.map toNode vs

------------------------------------------------------------------------
class ToNode a where
  toNode :: a -> ((Int, [Text]), [(Int,Int,Double)])

instance ToNode Csv_Col_5
  where
    toNode :: Csv_Col_5 -> ((Int, [Text]), [(Int,Int,Double)])
    toNode (Csv_Col_5 n v _ _ l) = (ln, es)
      where
        ln = (n, [l])
        es = List.zip3 (cycle [n]) ns (cycle [1])
        ns = map (read . unpack) $ splitOn "," v

instance ToNode Csv_Col_3
  where
    toNode :: Csv_Col_3 -> ((Int, [Text]), [(Int,Int,Double)])
    toNode (Csv_Col_3 n v l) = (ln, es)
      where
        ln = (n, map show [l])
        es = List.zip3 (cycle [n]) ns (cycle [1])
        ns = map (read . unpack) $ splitOn "," v

instance ToNode Csv_Col_2
  where
    toNode :: Csv_Col_2 -> ((Int, [Text]), [(Int,Int,Double)])
    toNode (Csv_Col_2 n v) = (ln, es)
      where
        ln = (n, [])
        es = List.zip3 (cycle [n]) ns (cycle [1])
        ns = map (read . unpack) $ splitOn "," v


toNodeUnlabelled :: Csv_Col_3 -> ((Int, ()), [(Int,Int,())])
toNodeUnlabelled (Csv_Col_3 n v l) = (ln, es)
  where
    ln = (n, ())
    es = List.zip3 (cycle [n]) ns (cycle [()])
    ns = map (read . unpack) $ splitOn "," v


------------------------------------------------------------------------
readFileCsvCillex :: FilePath -> IO (Header, Vector Csv_Col_5)
readFileCsvCillex = fmap readWith . BL.readFile

readFileCsv_Col_2 :: FilePath -> IO (Header, Vector Csv_Col_2)
readFileCsv_Col_2 = fmap readWith . BL.readFile

readFileCsv_Col_3 :: FilePath -> IO (Header, Vector Csv_Col_3)
readFileCsv_Col_3 = fmap readWith . BL.readFile


------------------------------------------------------------------------
csvDecodeOptions :: DecodeOptions
csvDecodeOptions = defaultDecodeOptions {decDelimiter = delimiter}
  where
    delimiter :: Word8
    delimiter = fromIntegral $ ord ','

readWith :: FromNamedRecord a => BL.ByteString -> (Header, Vector a)
readWith bs = case decodeByNameWith csvDecodeOptions bs of
      Left e        -> panic (pack e)
      Right csvDocs -> csvDocs

------------------------------------------------------------------------
data Csv_Col_2 =
  Csv_Col_2 { rc2_node    :: Int
           , rc2_voisins :: Text
           }
    deriving (Show)

instance FromNamedRecord Csv_Col_2 where
  parseNamedRecord r = Csv_Col_2 <$> r .: "@node: #id_pdg"
                                 <*> r .: "%+voisins"

---------------------------------------

data Csv_Col_3 =
  Csv_Col_3 { rc3_node    :: Int
            , rc3_voisins :: Text
            , rc3_id      :: Double
            }
    deriving (Show)

instance FromNamedRecord Csv_Col_3 where
  parseNamedRecord r =
    Csv_Col_3 <$> r .: "@node: #id_pdg"
              <*> r .: "%+voisins"
              <*> r .: "label"


---------------------------------------
data Csv_Col_5 =
  Csv_Col_5 { node    :: Int
            , voisins :: Text
            , lemme   :: Text
            , id      :: Double
            , label   :: Text
            }
    deriving (Show)

instance FromNamedRecord Csv_Col_5 where
  parseNamedRecord r = Csv_Col_5 <$> r .: "@node: #id_pdg"
                                 <*> r .: "%+voisins"
                                 <*> r .: "lemme"
                                 <*> r .: "id"
                                 <*> r .: "label"

{-
-- If needed to write down the CSV:
Csv_Col_5Header =
  header [ "@node: #id_pdg"
         , "%+voisins"
         , "id"
         , "lemme"
         , "label"
         ]

-}

